#########################################################################
## DATA PREPARATION #####################################################
#########################################################################
## download the hdm package
if (!require(hdm)) install.packages("hdm");
## include the hdm package
library(hdm);
## Prepare the BLP data
data(BLP);
BLPData <- BLP$BLP;
BLPData$price <- BLPData$price + 11.761
attach(BLPData);

## Baseline model (OLS & TSLS corresponding to Table III of BLP 1995)
Y <- as.matrix(y);
Y <- rbind(Y,NA);
X <- as.matrix(cbind(1, BLPData[,c("hpwt","air","mpd","space")]));
X <- rbind(X,NA);
D <- as.matrix(log(price));
D <- rbind(D,NA);
Z <- hdm:::constructIV(firm.id,cdid,id,X);
Z <- rbind(Z,NA);
I <- as.matrix(model.id);
J <- as.matrix(cdid);

## IJMatri stores the indices of data for i-th model (row) and j-th market (col)
N = max(I);
M = max(J);
IJMatrix = matrix(dim(Y)[1],N,M);
for(idx in 1:length(id)){
  IJMatrix[I[idx,],J[idx,]] = idx;
}
## IJ stores the idndices of data in a vectorized form
IJ = c(t(IJMatrix));

## Vectorized Y, X, D, and Z in the order of IJ indices (1,1),(1,2),(1,3),...
vecY = as.matrix(Y[IJ,]);
colnames(vecY) <- c("RelativeShare");
vecX = NULL;
for(idx in 1:(dim(X)[2])){
  vecX = cbind(vecX, as.matrix(c(X[IJ,idx])));
}
colnames(vecX) <- colnames(X);
vecD = as.matrix(D[IJ,]);
colnames(vecD) <- c("LogPrice");
vecZ = NULL;
for(idx in 1:(dim(Z)[2])){
  vecZ = cbind(vecZ, as.matrix(c(Z[IJ,idx])));
}
colnames(vecZ) <- colnames(Z);

## Make X high-dimensional
vecX = cbind(vecX,
             vecX[,2:5]^2,
             vecX[,2]*vecX[,3],vecX[,2]*vecX[,4],vecX[,2]*vecX[,5],
             vecX[,3]*vecX[,4],vecX[,3]*vecX[,5],
             vecX[,4]*vecX[,5],
             vecX[,2:5]^3,
             vecX[,2]^2*vecX[,3],vecX[,2]^2*vecX[,4],vecX[,2]^2*vecX[,5],
             vecX[,3]^2*vecX[,2],vecX[,3]^2*vecX[,2],vecX[,3]^2*vecX[,5],
             vecX[,4]^2*vecX[,3],vecX[,4]^2*vecX[,3],vecX[,4]^2*vecX[,5]);

#########################################################################
## DATA PREPARATION DONE ################################################
#########################################################################


#########################################################################
## ESTIMATION AND INFERENCE 2-WAY #######################################
#########################################################################
library(glmnet)
library(MASS)
K = 2;
Zcoordinate = 2;
Missing = is.na(vecY) | is.na(vecD) | is.na(vecZ[,Zcoordinate]);

#########################################################################
# ESTIMATION 2-WAY
#########################################################################
list_idx = matrix(1:N,N,1) %x% matrix(1,M,1);
list_jdx = matrix(1,N,1) %x% matrix(1:M,M,1);
k = ceiling(list_idx / ceiling(N/K));
l = ceiling(list_jdx / ceiling(M/K));
Y_Xbeta = NULL;
D_Xgamma = NULL;
Z_Xxi = NULL;
subsample_mean_Y_Xbeta_Z_Xxi = NULL;
subsample_mean_D_Xgamma_Z_Xxi = NULL;
for(kdx in 1:max(k)){
  for(ldx in 1:max(l)){
    ##############################################################
    # GET AUXILIARY SAMPLE (I^c X J^c)
    ##############################################################
    auxiliary_indices = which( k != kdx & l != ldx & !Missing);
    auxiliaryY = vecY[auxiliary_indices];
    auxiliaryD = vecD[auxiliary_indices];
    auxiliaryZ = vecZ[auxiliary_indices,Zcoordinate];
    auxiliaryX = vecX[auxiliary_indices,];
    
    ##############################################################
    # LASSO OF Y ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryY, standardize = TRUE, intercept = TRUE );
    mu1 = cvfit$lambda.min
    lasso1 = glmnet( auxiliaryX, auxiliaryY, lambda = mu1, standardize = TRUE, intercept = TRUE );
    beta_LASSO = coef( lasso1, s=mu1 );
    ##############################################################
    # LASSO OF D ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryD, standardize = TRUE, intercept = TRUE );
    mu2 = cvfit$lambda.min
    lasso2 = glmnet( auxiliaryX, auxiliaryD, lambda = mu2, standardize = TRUE, intercept = TRUE );
    gamma_LASSO = coef( lasso2, s=mu2 );
    ##############################################################
    # LASSO OF Z ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryZ, standardize = TRUE, intercept = TRUE );
    mu3 = cvfit$lambda.min
    lasso3 = glmnet( auxiliaryX, auxiliaryZ, lambda = mu3, standardize = TRUE, intercept = TRUE );
    xi_LASSO = coef( lasso3, s=mu3 );
    
    ##############################################################
    # GET MAIN SAMPLE (I X J)
    ##############################################################
    main_indices = which( k == kdx & l == ldx & !Missing);
    mainY = vecY[main_indices];
    mainD = vecD[main_indices];
    mainZ = vecZ[main_indices,Zcoordinate];
    mainX = matrix(vecX[main_indices,],length(main_indices),dim(vecX)[2]);
    
    ##############################################################
    # COMPUTE Y - Xbeta, D - Xgamma, AND Z - Xxi
    ##############################################################
    Y_Xbeta = rbind(Y_Xbeta, mainY - cbind(1,mainX) %*% beta_LASSO);
    D_Xgamma = rbind(D_Xgamma, mainD - cbind(1,mainX) %*% gamma_LASSO);
    Z_Xxi = rbind(Z_Xxi, mainZ - cbind(1,mainX) %*% xi_LASSO);
    
    ##############################################################
    # COMPUTE E(Y - Xbeta)*(Z - Xxi) AND E(D - Xgamma)*(Z - Xxi)
    ##############################################################
    subsample_mean_Y_Xbeta_Z_Xxi = rbind(subsample_mean_Y_Xbeta_Z_Xxi, mean((mainY - cbind(1,mainX) %*% beta_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
    subsample_mean_D_Xgamma_Z_Xxi = rbind(subsample_mean_D_Xgamma_Z_Xxi, mean((mainD - cbind(1,mainX) %*% gamma_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
  }
}

theta_hat_twoway = sum(subsample_mean_Y_Xbeta_Z_Xxi) / sum(subsample_mean_D_Xgamma_Z_Xxi);

#########################################################################
# VARIANCE ESTIMATION 2-WAY
#########################################################################
Epsilon_hat = Y_Xbeta - theta_hat_twoway * D_Xgamma;
temp = array(0,N*M);
temp[which(!Missing)] = Epsilon_hat;
Epsilon_hat = t(matrix(temp,M,N));

U_hat = D_Xgamma;
temp = array(0,N*M);
temp[which(!Missing)] = U_hat;
U_hat = t(matrix(temp,M,N));

V_hat = Z_Xxi;
temp = array(0,N*M);
temp[which(!Missing)] = V_hat;
V_hat = t(matrix(temp,M,N));

Gamma = 0;
J = 0;
for(kdx in 1:max(k)){
  for(ldx in 1:max(l)){
    ##############################################################
    # GET MAIN SAMPLE (I X J)
    ##############################################################
    main_indices = which( k == kdx & l == ldx );
    curr_list_idx = unique(list_idx[main_indices]);
    curr_list_jdx = unique(list_jdx[main_indices]);
    CardinalityIk = length(curr_list_idx);
    CardinalityJl = length(curr_list_jdx);
    
    for(idx1 in curr_list_idx){
      for(jdx1 in curr_list_jdx){
        
        for(jdx2 in curr_list_jdx){
          cardinality = CardinalityIk^2 * CardinalityJl^2 / min(CardinalityIk,CardinalityJl);
          Gamma = Gamma + cardinality^(-1) * Epsilon_hat[idx1,jdx1] * V_hat[idx1,jdx1] * Epsilon_hat[idx1,jdx2] * V_hat[idx1,jdx2];
        }#END FOR jdx2
        for(idx2 in curr_list_idx){
          cardinality = CardinalityIk^2 * CardinalityJl^2 / min(CardinalityIk,CardinalityJl);
          Gamma = Gamma + cardinality^(-1) * Epsilon_hat[idx1,jdx1] * V_hat[idx1,jdx1] * Epsilon_hat[idx2,jdx1] * V_hat[idx2,jdx1];
        }#END FOR idx2
        
        J = J - U_hat[idx1,jdx1] * V_hat[idx1,jdx1] / (CardinalityIk * CardinalityJl);
      }#END FOR idx1
    }#END FOR jdx1
    
    
  }#END FOR kdx
}#END FOR ldx

Gamma = Gamma / K^2;
J = J / K^2;
sigma2 = Gamma / J^2;
theta_SD_twoway = sqrt( sigma2 / min(N,M) );










#########################################################################
## ESTIMATION AND INFERENCE 0-WAY #######################################
#########################################################################
K = K^2;

#########################################################################
# ESTIMATION 0-WAY
#########################################################################
list_idx = 1:(dim(vecY)[1])
k = ceiling(list_idx / ceiling((dim(vecY)[1])/K));
Y_Xbeta = NULL;
D_Xgamma = NULL;
Z_Xxi = NULL;
subsample_mean_Y_Xbeta_Z_Xxi = NULL;
subsample_mean_D_Xgamma_Z_Xxi = NULL;
for(kdx in 1:max(k)){
    ##############################################################
    # GET AUXILIARY SAMPLE (I^c X J^c)
    ##############################################################
    auxiliary_indices = which( k != kdx & !Missing);
    auxiliaryY = vecY[auxiliary_indices];
    auxiliaryD = vecD[auxiliary_indices];
    auxiliaryZ = vecZ[auxiliary_indices,Zcoordinate];
    auxiliaryX = vecX[auxiliary_indices,];
    
    ##############################################################
    # LASSO OF Y ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryY, standardize = TRUE, intercept = TRUE );
    mu1 = cvfit$lambda.min
    lasso1 = glmnet( auxiliaryX, auxiliaryY, lambda = mu1, standardize = TRUE, intercept = TRUE );
    beta_LASSO = coef( lasso1, s=mu1 );
    ##############################################################
    # LASSO OF D ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryD, standardize = TRUE, intercept = TRUE );
    mu2 = cvfit$lambda.min
    lasso2 = glmnet( auxiliaryX, auxiliaryD, lambda = mu2, standardize = TRUE, intercept = TRUE );
    gamma_LASSO = coef( lasso2, s=mu2 );
    ##############################################################
    # LASSO OF Z ON X
    ##############################################################
    set.seed(1);
    cvfit = cv.glmnet( auxiliaryX, auxiliaryZ, standardize = TRUE, intercept = TRUE );
    mu3 = cvfit$lambda.min
    lasso3 = glmnet( auxiliaryX, auxiliaryZ, lambda = mu3, standardize = TRUE, intercept = TRUE );
    xi_LASSO = coef( lasso3, s=mu3 );
    
    ##############################################################
    # GET MAIN SAMPLE (I X J)
    ##############################################################
    main_indices = which( k == kdx & !Missing);
    mainY = vecY[main_indices];
    mainD = vecD[main_indices];
    mainZ = vecZ[main_indices,Zcoordinate];
    mainX = matrix(vecX[main_indices,],length(main_indices),dim(vecX)[2]);
    
    ##############################################################
    # COMPUTE Y - Xbeta, D - Xgamma, AND Z - Xxi
    ##############################################################
    Y_Xbeta = rbind(Y_Xbeta, mainY - cbind(1,mainX) %*% beta_LASSO);
    D_Xgamma = rbind(D_Xgamma, mainD - cbind(1,mainX) %*% gamma_LASSO);
    Z_Xxi = rbind(Z_Xxi, mainZ - cbind(1,mainX) %*% xi_LASSO);
    
    ##############################################################
    # COMPUTE E(Y - Xbeta)*(Z - Xxi) AND E(D - Xgamma)*(Z - Xxi)
    ##############################################################
    subsample_mean_Y_Xbeta_Z_Xxi = rbind(subsample_mean_Y_Xbeta_Z_Xxi, mean((mainY - cbind(1,mainX) %*% beta_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
    subsample_mean_D_Xgamma_Z_Xxi = rbind(subsample_mean_D_Xgamma_Z_Xxi, mean((mainD - cbind(1,mainX) %*% gamma_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
}

theta_hat_zeroway = sum(subsample_mean_Y_Xbeta_Z_Xxi) / sum(subsample_mean_D_Xgamma_Z_Xxi);

#########################################################################
# VARIANCE ESTIMATION 0-WAY
#########################################################################
Epsilon_hat = Y_Xbeta - theta_hat_twoway * D_Xgamma;
temp = array(0,N*M);
temp[which(!Missing)] = Epsilon_hat;
Epsilon_hat = t(matrix(temp,1,M*N));

U_hat = D_Xgamma;
temp = array(0,N*M);
temp[which(!Missing)] = U_hat;
U_hat = t(matrix(temp,1,M*N));

V_hat = Z_Xxi;
temp = array(0,N*M);
temp[which(!Missing)] = V_hat;
V_hat = t(matrix(temp,1,M*N));

Gamma = 0;
J = 0;
for(kdx in 1:max(k)){
    ##############################################################
    # GET MAIN SAMPLE (I X J)
    ##############################################################
    main_indices = which( k == kdx );
    curr_list_idx = unique(list_idx[main_indices]);
    CardinalityIk = length(curr_list_idx);
    
    for(idx1 in curr_list_idx){

        for(idx2 in curr_list_idx){
          Gamma = Gamma + CardinalityIk^(-1) * Epsilon_hat[idx1,1] * V_hat[idx1,1] * Epsilon_hat[idx2,1] * V_hat[idx2,1];
        }#END FOR idx2
        
        J = J - U_hat[idx1,1] * V_hat[idx1,1] / CardinalityIk;
    }#END FOR idx1
    
    
}#END FOR kdx

Gamma = Gamma / K;
J = J / K;
sigma2 = Gamma / J;
theta_SD_zeroway = sqrt( sigma2 / sum(!Missing) );












#########################################################################
## ESTIMATION AND INFERENCE IV LASSO (0-WAY) ############################
#########################################################################
lasso.reg <- rlassoIV(x=vecX[!Missing,-1], y=vecY[!Missing,1], z=vecZ[!Missing,Zcoordinate], d=vecD[!Missing,1], select.X=TRUE, select.Z=TRUE, intercept = TRUE)









#########################################################################
## ESTIMATION AND INFERENCE IV LASSO (0-WAY) ############################
#########################################################################
rbind(theta_hat_twoway,theta_SD_twoway)
rbind(theta_hat_zeroway,theta_SD_zeroway)
summary(lasso.reg)

